;;;
;;; This file defines the basic infrastructure for the bindings.
;;; I.e., some macros for easer maintance are defined here and
;;; the library itself is loaded.
;;;

(in-package :cffi-cairo)

(define-foreign-library libcairo
  (:unix (:or "libcairo.so.2" "libcairo.so"))
  (t (:default "libcairo")))

(use-foreign-library libcairo)

(defun make-fun-symbol (c-name)
  "creates a symbol name according to a c (function) name"
  (intern (substitute #\- #\_ (string-upcase c-name))))

(defun make-cairo-fun-name (name)
  "creates a cairo funtion name out of a given name.
   (Basically a convinence method to prepend cairo_.)"
  (concatenate 'string "cairo_" name))

(defmacro defcairofun (fun &body options)
  "Creates a cairo function binding for fun according to the options,
   where options can specify the following things:
   :return-type  The return type of the function.
   :args         The arguments of the function
   :alternative  An alternative (more lispy version) for the function
                 (described later)."
  (let ((return-type (cadr (assoc :return-type options)))
	(args        (cdr (assoc :args options)))
	(alternative (cdr (assoc :alternative options))))
    (let ((return-type* (if (null return-type) :void return-type))
	  (c-fun-name   (make-cairo-fun-name (if (consp fun) (car fun) fun)))
	  (fun-symbol   (if (consp fun) (cadr fun) (make-fun-symbol fun))))
      `(export
	(list
	 ,@(cons `(defcfun (,c-fun-name ,fun-symbol) ,return-type* ,@args)
		 (if (null alternative)
		     nil
		     (cons
		      `(defcairo-alternative ,fun-symbol ,args
			,(cdr (assoc :args alternative))
			,(cadr (assoc :orig-value alternative))
			,(cdr (assoc :pointers alternative))
			,@(cdr (assoc :body alternative)))
		      nil))))))))

(defun class-alternative (name alternative)
  (if (null alternative)
      nil
      `((:alternative
	 (:args ,name ,@(cdr (assoc :args alternative)))
	 ,@alternative))))

(defun class-fun-name (class fun)
  (if (null class)
      fun
      (concatenate 'string class "_" fun)))

(defmacro defcairoclass (class name type &body defs)
  "Macro to define a class of cairo function.
   This assembles a group of defcairofun definitions that allows
   to ommit the class / group name, and part of the function
   signature for even shorter binding specifications.
   Cairo functions are organized in groups and the function
   name allways have the form cairo_[group]_... and either
   the first argument is of a type specific to that group or
   the function creates an object of that type.
   Hence, you define a list of constructors and methods,
   where just the function name is specified and the
   return-type of constructors and the first argument of methods
   is ommited."
  `(progn
    ,@(loop :for def :in (cdr (assoc :constructors defs))
	    :for (fun . rest) = def
	    :for fun* = (class-fun-name class fun)
	    :for rest* = (cons `(:return-type ,type) rest)
	    :collect `(defcairofun ,@(cons fun* rest*)))
    ,@(loop :for def :in (cdr (assoc :methods defs))
	    :for (fun . rest) = def
	    :for fun* = (class-fun-name class fun)
	    :for alternative = (cdr (assoc :alternative (cdr def)))
	    :for rest* = `((:args (,name ,type)
			    ,@(cdr (assoc :args rest)))
			   ,@(class-alternative name alternative)
			   ,@rest)
	    :collect `(defcairofun ,@(cons fun* rest*)))))

(defun make-alt-symbol (fun)
  "Creates a symbol for an alternative to fun."
  (intern (concatenate 'string (symbol-name fun) "*")))

(defun make-ref-pointer-fun (p-syms)
  (lambda (pointer)
    (let ((pointer-sym  (car pointer))
	  (pointer-type (cadr pointer)))
      (let ((p-sym (cdr (assoc pointer-sym p-syms))))
	`(,(make-alt-symbol pointer-sym)
	  (mem-ref ,(if (null p-sym) pointer-sym p-sym) ,pointer-type))))))

(defun make-foreign-objects-fun (p-syms)
  (lambda (p)
    (let ((p-sym (cdr (assoc (car p) p-syms))))
      (cons (if (null p-sym) (car p) p-sym) (cdr p)))))

(defun make-map-argument (p-syms)
  (lambda (p)
    (let ((p-sym (cdr (assoc p p-syms))))
      (if (null p-sym) p p-sym))))

(defmacro defcairo-alternative (fun orig-args alt-args ret pointers &body body)
  "Create an alternative function for a function binding.
   orig-args  The original argument list (list of name / type pairs).
   alt-args   The alternative argument list (list of names)
   ret        Symbol to bind the return value to, may be nil.
   pointers   List of pairs (name, type) of pointer arguments.
              For every pointer with symbol s, a symbol s* is
              bound to the corresponding value the pointer points
              to in the alternative function body.
   body       The function body of the alternative function."
  (let* ((pointers*       (mapcar #'car pointers))
	 (orig-args*      (mapcar #'car orig-args))
	 (in-out          (intersection pointers* alt-args))
	 (p-syms          (mapcar #'(lambda (p) (cons p (gensym))) in-out))
	 (orig-fun        (if (consp fun) (car fun) fun))
	 (mapd-orig-args  (mapcar (make-map-argument p-syms) orig-args*))
	 (alt-fun         (if (consp fun) (cadr fun) (make-alt-symbol fun)))
	 (foreign-objects (mapcar (make-foreign-objects-fun p-syms)
				  pointers))
	 (pointer-refs    (mapcar (make-ref-pointer-fun p-syms) pointers))
	 (setf-pointers   (mapcar #'(lambda (p)
				      `(setf (mem-ref ,(cdr (assoc p p-syms))
					      ,(cadr (assoc p pointers))) ,p))
				  in-out)))
    `(defun ,alt-fun ,alt-args
      (with-foreign-objects ,foreign-objects
	,@setf-pointers
	,(if (not (null ret))
	     `(let ((,ret (,orig-fun ,@mapd-orig-args)))
	       (symbol-macrolet ,pointer-refs ,@body))
	     `(progn (,orig-fun ,@mapd-orig-args)
	       (symbol-macrolet ,pointer-refs ,@body)))))))

